#include <stdio.h>
#include "strerr.h"
#include "sml.h"
#include "stack.h"
#include "optimize.h"
#include "permute.h"
#include "str.h"
#include "timeout.h"

static ContextStack *
find_best_transition(ContextStack *top,
    int *costp, stralloc *rtagsp, int lookahead, Run *run);

static int
try_outputb(const char *b, int n)
{
    if (new_output->len + n >= safe_output->len)
	return 0;

    stralloc_catb(new_output, b, n);
    return 1;
}

static int
try_outputs(const char *s)
{
    return try_outputb(s, str_len(s));
}

static ContextStack *
pop_one_tag(ContextStack *top, stralloc *tags)
{
    if (tags && !stralloc_cats(tags, tag_close[(int) top->x.tag]))
	quit();

    return top->next;
}

static ContextStack *
push_one_tag(ContextStack *top, TagTypes tag, int *costp, stralloc *tags)
{
    if (tags && !stralloc_cats(tags, tag_open[tag]))
	quit();
    *costp += tag_cost[tag];
    return ContextStack_push(top, tag);
}

static ContextStack *
push_tags(ContextStack *top, char *types, int n, int *costp, stralloc *tags)
{
    while (n) {
	top = push_one_tag(top, *types, costp, tags);
	types++;
	n--;
    }
    return top;
}

static int
need_pops_or_pl(char topflags, char goalflags)
{
    return (topflags & (B|I|TT|S) & ~goalflags)
	|| ((topflags & EM) && !(goalflags & (EM|S)))
	|| ((topflags & U) > (goalflags & U));
}

#define MAX_TAGS 10

static ContextStack *
find_best_transition_pushes_only(ContextStack *top,
    int *costp, stralloc *rtagsp, int lookahead, Run *run)
{
    char goalflags = run->x.flags;
    char topflags = top->x.flags;
    char bits = goalflags & ~topflags;
    int udiff = (goalflags & U) - (topflags & U);
    char types[MAX_TAGS];
    char *t = types;
    int ntags;

    if (udiff > 0) {
	int i = udiff;
	while (i > 0) {
	    *t++ = Tag_U;
	    i--;
	}
    }
    if (bits & B) *t++ = Tag_B;
    if (bits & I) *t++ = Tag_I;
    if (bits & TT) *t++ = Tag_TT;
    if (bits & S) *t++ = Tag_S;
    if (bits & EM) *t++ = Tag_EM;
    if (run->x.size != top->x.size)
	*t++ = Tag_0 + run->x.size - '0';
    if (run->x.color != top->x.color)
	*t++ = TagType_for_color(run->x.color);

    ntags = t - types;

    if (lookahead > 1 && ntags > 1) {
	int trycost;
	ContextStack *trytop;
	stralloc trytags = { 0 };
	stralloc *trytagsp = rtagsp ? &trytags : 0;

	int bestcost = 0x7fffffff;
	ContextStack *besttop;
	stralloc besttags = { 0 };

	int nperms = npermutations(ntags, udiff);
	int i;
	char ptypes[MAX_TAGS];

	/* !!! consider not trying all permutations */
	for (i = 0; i < nperms; i++) {
	    permute(ptypes, types, ntags, i, udiff);
	    trycost = 0;
	    trytags.len = 0;
	    trytop = push_tags(top, ptypes, ntags, &trycost, trytagsp);

	    find_best_transition(trytop,
		&trycost, 0, lookahead - 1, run + 1);

	    if (trycost < bestcost) {
		bestcost = trycost;
		besttop = trytop;
		if (trytagsp) stralloc_copy(&besttags, &trytags);
	    }
	}

	if (rtagsp) {
	    stralloc_cat(rtagsp, &besttags);
	    stralloc_unready(&trytags);
	    stralloc_unready(&besttags);
	}

	*costp += bestcost;
	return besttop;
    }

    return push_tags(top, types, t - types, costp, rtagsp);
}

static ContextStack *
find_best_transition_optional_pops(ContextStack *top,
    int *costp, stralloc *rtagsp, int lookahead, Run *run)
{
    int bestcost = 0x7fffffff;
    stralloc besttags = { 0 };
    ContextStack *besttop;

    int trycost;
    stralloc trytags = { 0 };
    stralloc *trytagsp = rtagsp ? &trytags : 0;
    ContextStack *trytop;
    int poppedtagslen;

    trytags.len = 0;

    while(1) {
	poppedtagslen = trytags.len;

	if (!need_pops_or_pl(top->x.flags, run->x.flags)) {
	    trycost = 0;
	    trytop = find_best_transition_pushes_only(top,
		&trycost, trytagsp, lookahead, run);

	    if (trycost < bestcost) {
		bestcost = trycost;
		besttop = trytop;
		if (trytagsp) stralloc_copy(&besttags, &trytags);
	    }
	}

	if (!top->x.tag) break;

	trytags.len = poppedtagslen;
	top = pop_one_tag(top, &trytags);
    }

    if (rtagsp) {
	stralloc_cat(rtagsp, &besttags);
	stralloc_unready(&trytags);
	stralloc_unready(&besttags);
    }

    *costp += bestcost;

    return besttop;
}

static ContextStack *
find_best_transition_font_popped(ContextStack *top,
    int *costp, stralloc *rtagsp, int lookahead, Run *run)
{
    char goalflags = run->x.flags;

    if (need_pops_or_pl(top->x.flags, goalflags)) {

	int plcost = 0;
	stralloc pltags = { 0 };
	stralloc *pltagsp = rtagsp ? &pltags : 0;
	ContextStack *pltop;

	int popcost = 0;
	stralloc poptags = { 0 };
	stralloc *poptagsp = rtagsp ? &poptags : 0;
	ContextStack *poptop;

	/* Try pushing PL. */

	pltop = push_one_tag(top, Tag_PL, &plcost, pltagsp);
	pltop = find_best_transition_pushes_only(pltop,
	    &plcost, pltagsp, lookahead, run);

	/* Try popping tags. */

	poptop = top;
	popcost = 0;
	do {
	    poptop = pop_one_tag(poptop, poptagsp);
	} while (need_pops_or_pl(poptop->x.flags, goalflags));

	poptop = find_best_transition_optional_pops(poptop,
	    &popcost, poptagsp, lookahead, run);

	/* Use the best. */

	/* !!! Consider testing (popcost <= plcost) */
	if (popcost < plcost) {
	    *costp += popcost;
	    if (rtagsp) stralloc_cat(rtagsp, &poptags);
	    top = poptop;
	}

	else {
	    *costp += plcost;
	    if (rtagsp) stralloc_cat(rtagsp, &pltags);
	    top = pltop;
	}

	if (rtagsp) {
	    stralloc_unready(&pltags);
	    stralloc_unready(&poptags);
	}

	return top;
    }

    return find_best_transition_optional_pops(top, costp, rtagsp,
	lookahead, run);
}

static ContextStack *
find_best_transition(ContextStack *top,
    int *costp, stralloc *rtagsp, int lookahead, Run *run)
{
    if (!run->x.size) {
	while (top->x.size)
	    top = pop_one_tag(top, rtagsp);
    }

    if (!run->x.color) {
	while (top->x.color)
	    top = pop_one_tag(top, rtagsp);
    }

    return find_best_transition_font_popped(top, costp, rtagsp,
	lookahead, run);

}

/* returns 0 if new_output would be longer than safe_output */
static int
clear_stack(ContextStack *top)
{
    while (top->x.tag) {
	if (!try_outputs(tag_close[(int) top->x.tag]))
	    return 0;
	top = top->next;
    }
    return 1;
}

void
optimize(int lookahead)
{
    char *p = chars.s;
    int runsleft = markup.len;
    Run *run = markup.s;
    int abort = 0;
    ContextStack *top;
    stralloc tags = { 0 };

    top = ContextStack_gc(bottom);

    /* First run is always outside of all tags. */
    stralloc_copyb(new_output, p, run->len);
    p += run->len;

    for (--runsleft, ++run; runsleft; --runsleft, ++run) {
	int cost = 0;
	tags.len = 0;

	if (lookahead > runsleft) lookahead = runsleft;

	top = find_best_transition(top, &cost, &tags, lookahead, run);

	if (0
	    || !try_outputb(tags.s, tags.len)
	    || !try_outputb(p, run->len)
	) {
	    abort = 1;
	    break;
	}

	p += run->len;
	top = ContextStack_gc(top);
    }

    stralloc_unready(&tags);
    if (abort || !clear_stack(top)) {
/*	strerr_warn1("not flipping", 0);*/
	return;
    }
/*    strerr_warn1("flipping", 0);*/
    flip_outputs();
}

